import requests
import ssl

from requests.adapters import HTTPAdapter
from urllib3.poolmanager import PoolManager


class HTTPtimeoutAdapter(HTTPAdapter):
    """ Adds timeout to requests HTTP adapter """

    def __init__(self, timeout: int = 5, max_retries: int = 0, *args, **kwargs):
        super().__init__(*args, **kwargs, max_retries=max_retries)
        self.timeout = timeout

    def init_poolmanager(self, connections, maxsize, block=False, **pool_kwargs):
        """Initializes a urllib3 PoolManager.

        This method should not be called from user code, and is only
        exposed for use when subclassing the
        :class:`HTTPAdapter <requests.adapters.HTTPAdapter>`.

        :param connections: The number of urllib3 connection pools to cache.
        :param maxsize: The maximum number of connections to save in the pool.
        :param block: Block when no free connections are available.
        :param pool_kwargs: Extra keyword arguments used to initialize the Pool Manager.
        """
        # save these values for pickling
        self._pool_connections = connections
        self._pool_maxsize = maxsize
        self._pool_block = block

        self.poolmanager = PoolManager(num_pools=connections, maxsize=maxsize,
                                       block=block, strict=True, ssl_version=ssl.PROTOCOL_TLSv1_2,
                                       **pool_kwargs)

    def send(self, *args, **kwargs):
        if kwargs.get('timeout') is None:
            kwargs['timeout'] = self.timeout
        return super(HTTPtimeoutAdapter, self).send(*args, **kwargs)


class SunshineBase:
    _OBJECTS_ROOT = 'api/sunshine/objects'

    _RELATIONSHIPS_ROOT = 'api/sunshine/relationships'

    _PROFILES_ROOT = 'api/v2/user_profiles'

    _EVENTS_ROOT = 'api/v2/user_profiles'

    _USERS_ROOT = 'api/v2/users'

    _HEADERS = {
        'Content-type': 'application/json',
        'Accept': 'application/json'
    }

    def __init__(self, subdomain: str, email: str, key: str, global_timeout: int = 5, max_retries: int = 0):
        """
        Creates and manages objects on zendesk sunshine

        :param subdomain: Your zendesk subdomain
        :param email: Sunshine user email ID. Used for token based authentication
        :param key: Sunshine token.
        :param global_timeout: Request timeout in seconds for every request within the session
        :param max_retries: The maximum number of retries each connection
            should attempt. Note, this applies only to failed DNS lookups, socket
            connections and connection timeouts, never to requests where data has
            made it to the server. By default, Requests does not retry failed
            connections. If you need granular control over the conditions under
            which we retry a request, import urllib3's ``Retry`` class and pass
            that instead.
        """

        self.subdomain = subdomain
        self._objects_base_url = f'https://{self.subdomain}.zendesk.com/{self._OBJECTS_ROOT}'
        self._relationships_base_url = f'https://{self.subdomain}.zendesk.com/{self._RELATIONSHIPS_ROOT}'
        self._profiles_base_url = f'https://{self.subdomain}.zendesk.com/{self._PROFILES_ROOT}'
        self._events_base_url = f'https://{self.subdomain}.zendesk.com/{self._EVENTS_ROOT}'
        self._users_base_url = f'https://{self.subdomain}.zendesk.com/{self._USERS_ROOT}'

        self._session = requests.Session()

        self._adapter = HTTPtimeoutAdapter(global_timeout, max_retries=max_retries)
        self._session.mount('http://', self._adapter)
        self._session.mount('https://', self._adapter)

        self._session.headers.update(self._HEADERS)

        self._session.auth = (f'{email}/token', f'{key}')

    def __repr__(self):
        return f'{self.__class__.__name__}(subdomain="{self.subdomain}")'

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._session.close()

    def close_session(self):
        """ Closes requests session """

        self._session.close()



"""
Zendesk sunshine API
API guide: https://developer.zendesk.com/rest_api/docs/sunshine/introduction
"""
from hns_sunshine_api.sunshine.object_records import SunshineObjectRecords
from hns_sunshine_api.sunshine.relationship_records import SunshineRelationshipRecords
from hns_sunshine_api.sunshine.profiles import SunshineProfiles
from hns_sunshine_api.sunshine.events import SunshineEvents
from hns_sunshine_api.sunshine.base import SunshineBase


class Sunshine(SunshineBase):

    def __init__(self, subdomain: str, email: str, key: str, global_timeout: int = 5, max_retries: int = 0):
        """
        Creates and manages objects on zendesk sunshine

        Instantiating creates a requests session. You can call `close` at the end.

        Usage:

        Simple use:

            sunshine = Sunshine(subdomain='mydomain', email='test@example.com', key='test_key')
            objects = sunshine.object_records().list(object_type='sometype')
            relation = sunshine.relationships_records().show(relationship_id='someid')
            sunshine.close_session()

        Using context manager:

            with Sunshine(subdomain='mydomain', email='test@example.com', key='test_key') as sunshine:
                objects = sunshine.object_records().list(object_type='sometype')
                relation = sunshine.relationships_records().show(relationship_id='someid')

        Zendesk API docs: https://developer.zendesk.com/rest_api/docs/sunshine/introduction

        :param subdomain: Your zendesk subdomain
        :param email: Sunshine user email ID. Used for token based authentication
        :param key: Sunshine token.
        :param global_timeout: Request timeout in seconds for every request within the session
        :param max_retries: The maximum number of retries each connection
            should attempt. Note, this applies only to failed DNS lookups, socket
            connections and connection timeouts, never to requests where data has
            made it to the server. By default, Requests does not retry failed
            connections. If you need granular control over the conditions under
            which we retry a request, import urllib3's ``Retry`` class and pass
            that instead.
        """

        super().__init__(subdomain, email, key, global_timeout, max_retries=max_retries)
        self._object_record = SunshineObjectRecords(
            subdomain=subdomain, email=email, key=key, global_timeout=global_timeout, max_retries=max_retries
        )
        self._relationship_record = SunshineRelationshipRecords(
            subdomain=subdomain, email=email, key=key, global_timeout=global_timeout, max_retries=max_retries
        )
        self._profiles = SunshineProfiles(
            subdomain=subdomain, email=email, key=key, global_timeout=global_timeout, max_retries=max_retries
        )
        self._events = SunshineEvents(
            subdomain=subdomain, email=email, key=key, global_timeout=global_timeout, max_retries=max_retries
        )

    def __repr__(self):
        return f'{self.__class__.__name__}(subdomain="{self.subdomain}")'

    def object_records(self):
        return self._object_record

    def relationships_records(self):
        return self._relationship_record

    def object_types(self):
        raise NotImplementedError

    def relationship_types(self):
        raise NotImplementedError

    def events(self):
        return self._events

    def profiles(self):
        return self._profiles




from hns_sunshine_api.sunshine.base import SunshineBase
from requests import Response


class SunshineObjectRecords(SunshineBase):
    """ Sunshine Objects class """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __repr__(self):
        return f'{self.__class__.__name__}(subdomain="{self.subdomain}")'

    def list(self, object_type: str, external_id: str = None, timeout: int = None) -> Response:
        """
        Gets object records from sunshine
        :param object_type: Type of object record
        :param external_id: external ID for the object record
        :param timeout: Request timeout in seconds
        :return: Object record
        """

        params = {
            'type': object_type
        }
        if external_id:
            params['external_id'] = external_id

        return self._session.get(
            f'{self._objects_base_url}/records',
            params=params,
            timeout=timeout
        )

    def list_related(self, object_id: str, relationship_key: str, timeout: int = None) -> Response:
        """
        Returns all the object records that the specified object record has relationship records with for the
        specified relationship type
        :param object_id: Object record ID
        :param relationship_key: Relationship type key
        :param timeout: Request timeout in seconds
        :return: All the object records related to the relationship key
        """

        return self._session.get(
            f'{self._objects_base_url}/records/{object_id}/related/{relationship_key}',
            timeout=timeout
        )

    def show(self, object_id: str, timeout: int = None) -> Response:
        """
        Returns the specified object record
        :param object_id: Object record ID
        :param timeout: Request timeout in seconds
        :return: Object record
        """

        return self._session.get(
            f'{self._objects_base_url}/records/{object_id}',
            timeout=timeout
        )

    def create(self, record_data: dict, timeout: int = None) -> Response:
        """
        Creates an object record.
        :param record_data: Object record data.
        Check https://developer.zendesk.com/rest_api/docs/sunshine/resources#json-format for details
        :param timeout: Request timeout in seconds
        :return: Object record
        """

        return self._session.post(
            f'{self._objects_base_url}/records',
            json=record_data,
            timeout=timeout
        )

    def update(self, object_id: str, data: dict, timeout: int = None) -> Response:
        """
        Updates the attributes object of the specified object record. It does not update any other record properties.

        The attributes object patches the previously stored object. Therefore, the request should only contain the
        properties of the attributes object that need to be updated.

        The request must include an "application/merge-patch+json" content-type header.

        :param object_id: Object record ID
        :param data: Object record
        :param timeout: Request timeout in seconds
        :return: Updated object record
        """

        self._session.headers['Content-type'] = 'application/merge-patch+json'
        return self._session.patch(
            f'{self._objects_base_url}/records/{object_id}',
            json=data,
            timeout=timeout
        )

    def update_by_external_id(self, data: dict, timeout: int = None) -> Response:
        """
        Creates a new object if an object with given external id does not exist and updates the attributes object of
        the specified object record if an object with the given external id does exist.
        This endpoint does not update any other record properties.

        The request data should contain:
            - Object type
            - External id
            - attributes of the object that needs to be updated
        The request must include an "application/merge-patch+json" content-type header.

        :param data: Object record
        :param timeout: Request timeout in seconds
        :return: Updated object record
        """

        self._session.headers['Content-type'] = 'application/merge-patch+json'
        return self._session.patch(
            f'{self._objects_base_url}/records',
            json=data,
            timeout=timeout
        )

    def delete(self, object_id: str, timeout: int = None) -> Response:
        """
        Deletes the specified object record.

        Before deleting an object record, you must delete any relationship record that specifies the object record

        :param object_id: Object record ID
        :param timeout: Request timeout in seconds
        :return: Nothing
        """

        return self._session.delete(
            f'{self._objects_base_url}/records/{object_id}',
            timeout=timeout
        )



from hns_sunshine_api.sunshine.base import SunshineBase
from requests import Response


class SunshineRelationshipRecords(SunshineBase):
    """ Sunshine relationships class """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __repr__(self):
        return f'{self.__class__.__name__}(subdomain="{self.subdomain}")'

    def list_by_object_record(self, object_id: str, relationship_type: str, timeout: int = None) -> Response:
        """
        Returns the relationship records of a specified relationship type for a specified object record.
        For the type, specify a relationship type key. For the object record id, specify a custom object
        record id or a Zendesk object record id.
        The record id must be the relationship record's source record id, not its target record id.

        :param object_id: Object record ID
        :param relationship_type: Relationship type key
        :param timeout: Request timeout
        :return: Relationship record
        """

        return self._session.get(
            f'{self._objects_base_url}/records/{object_id}/relationships/{relationship_type}',
            timeout=timeout
        )

    def list_by_type(self, relationship_type: str, timeout: int = None) -> Response:
        """
        Returns all the relationship records of the specified relationship type.
        For the type, specify a relationship type key.
        :param relationship_type: Relationship type key
        :param timeout: Request timeout
        :return: Relationship records
        """

        return self._session.get(
            f'{self._relationships_base_url}/records',
            params={'type': relationship_type},
            timeout=timeout
        )

    def show(self, relationship_id: str, timeout: int = None) -> Response:
        """
        Returns a relationship record specified by id
        :param relationship_id: Relationship type key
        :param timeout: Request timeout
        :return: Relationship record
        """

        return self._session.get(
            f'{self._relationships_base_url}/records/{relationship_id}',
            timeout=timeout
        )

    def create(self, record_data: dict, timeout: int = None) -> Response:
        """
        Creates a relationship record between two object records based on a relationship type.
        The relationship type defines the object types of the two object records.
        :param record_data: Relationship record data.
        Check https://developer.zendesk.com/rest_api/docs/sunshine/relationships#json-format for details
        :param timeout: Request timeout
        :return: Relationship record
        """

        return self._session.post(
            f'{self._relationships_base_url}/records',
            json=record_data,
            timeout=timeout
        )

    def delete(self, relationship_id: str, timeout: int = None) -> Response:
        """
        Deletes a specified relationship record
        :param relationship_id: Relationship object ID
        :param timeout: Request timeout
        :return: Nothing
        """

        return self._session.delete(
            f'{self._relationships_base_url}/records/{relationship_id}',
            timeout=timeout
        )




from hns_sunshine_api.sunshine.base import SunshineBase
from requests import Response
from typing import Union, Dict


class SunshineProfiles(SunshineBase):
    """ Sunshine Profiles class """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __repr__(self):
        return f'{self.__class__.__name__}(subdomain="{self.subdomain}")'

    def get_by_identifier(self, identifier: str) -> Response:
        """
        Gets Sunshine profile by identifier
        :param identifier: Profile identifier. https://develop.zendesk.com/hc/en-us/articles/360043415094
        :return: Profile
        """

        return self._session.get(self._profiles_base_url, params={'identifier': identifier})

    def get_by_id(self, profile_id: str) -> Response:
        """
        Gets Sunshine profile by profile ID
        :param profile_id: Sunshine profile ID
        :return: Profile
        """

        return self._session.get(f'{self._profiles_base_url}/{profile_id}')

    def get_by_user_id(self, user_id: str) -> Response:
        """
        Gets Sunshine profile by Zendesk user ID
        :param user_id: Zendesk user ID
        :return: Profiles
        """

        return self._session.get(f'{self._users_base_url}/{user_id}/profiles')

    def profile_exists(self, identifier: str) -> Union[Dict[str, str], None]:
        """
        Checks is a Sunshine profile exists for a given identifier.
        :param identifier: Profile identifier. https://develop.zendesk.com/hc/en-us/articles/360043415094
        :return: If profile exists then returns a dict with keys: user_id (zendesk user id) and profile_id
        else returns None
        """

        profile = self.get_by_identifier(identifier)
        if profile.ok:
            profile_data = profile.json()
            return {'profile_id': profile_data['profile']['id'], 'user_id': profile_data['profile']['user_id']}
        return None

    def create_or_update(self, identifier: str, profile_data: dict) -> Response:
        """
        Creates or updates a profile
        :param identifier: Profile identifier. https://develop.zendesk.com/hc/en-us/articles/360043415094
        :param profile_data: Profile data. https://develop.zendesk.com/hc/en-us/articles/360044532633
        :return: Profile
        """

        return self._session.put(
            self._profiles_base_url,
            params={'identifier': identifier},
            json=profile_data
        )

    def partial_update(self, identifier: str, profile_data: dict) -> Response:
        """
        Partially updates a profile. If the profile does not exists, 404 error is returned.
        https://develop.zendesk.com/hc/en-us/articles/360044129314#patch
        :param identifier: Profile identifier. https://develop.zendesk.com/hc/en-us/articles/360043415094
        :param profile_data: Profile data. https://develop.zendesk.com/hc/en-us/articles/360044532633
        :return: Profile
        """

        return self._session.patch(
            self._profiles_base_url,
            params={'identifier': identifier},
            json=profile_data
        )

    def update_by_id(self, profile_id: str, profile_data: dict) -> Response:
        """
        Updates a profile by profile ID
        :param profile_id: Sunshine profile ID
        :param profile_data: Profile data. https://develop.zendesk.com/hc/en-us/articles/360044532633
        :return: Profile
        """

        return self._session.put(
            f'{self._profiles_base_url}/{profile_id}',
            json=profile_data
        )

    def partial_update_by_id(self, profile_id: str, profile_data: dict) -> Response:
        """
        Partially updates a profile by profile ID. If the profile does not exists, 404 error is returned.
        https://develop.zendesk.com/hc/en-us/articles/360044129314#patch
        :param profile_id: Sunshine profile ID
        :param profile_data: Profile data. https://develop.zendesk.com/hc/en-us/articles/360044532633
        :return: Profile
        """

        return self._session.patch(
            f'{self._profiles_base_url}/{profile_id}',
            json=profile_data
        )

    def delete_profile(self, profile_id: str) -> Response:
        """
        Deletes a Sunshine profile by profile ID
        :param profile_id: Sunshine profile ID
        :return: Just the HTTP status code
        """

        return self._session.delete(f'{self._profiles_base_url}/{profile_id}')
